/************************************************************************

    MikMod/2 Player - By Kevin Langman

    This is my crappy (Quick & Dirty) interface to the MIKMOD Sound Library.. 

    Any and all credits should go to the authors of the Library!! All I did was
    this simple interface. All of 600 lines compared to the countless lines of 
    code that was required to create the library!

    ----------------------------------------------------------------------
	This code is free software; you can redistribute it and/or modify
	it under the terms of the GNU Library General Public License as
	published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.
 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Library General Public License for more details.
 
	You should have received a copy of the GNU Library General Public
	License along with this code; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA.


************************************************************************/

#define true  1
#define false 0

#define MIKMOD2_VER "1.0"

#define NONE 0
#define EXIT 1
#define NEXT 2
#define PREV 3

#define INCL_BASE
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <sys/kbdscan.h>
     
#include <mikmod.h>

typedef struct _SongList{
    char             *pszFileName;
    struct _SongList *Next;
    struct _SongList *Prev;
} SongList;

int         HandleInput();
int         ParceArgs(int, char**, SongList **);
int         ParceEnv(char *);
void        DrawInstruments(MODULE*);
void        DrawSamples(MODULE*);
int         PlayModule(SongList **);
SongList*   LoadListFile(char *, SongList **);
int         readKBD() {  return( _read_kbd(0,0,0)); }

int volume = 128;
int wrap   = 0;
int loop   = 1;
int setPriority = true;

//static char szNoShareEnv[]="MM_NOSHARE=YES";

/************************************************************ MAIN ***********/
int main(int argc, char *argv[]){
    long     rc;
    char     *pszFileName;
    SongList *ModFiles;

    printf( "MikMod/2 - OS/2 Module Player using the MikMod sound library! - By Kevin Langman\n" );

    /* Set the Default settings */
    md_mode    = DMODE_SOFT_MUSIC | DMODE_16BITS | DMODE_STEREO;
    md_device  = 1;
    md_mixfreq = 44100;
    md_reverb  = 0;

    if( ParceEnv(getenv("MIKMOD2"))!=true){
        printf( "There is an error in the MIKMOD2 Enviroment string.\n" );
    }

    /* Parce Command Line Arguments and Ensure all required input has been supplied */
    if( argc==1 || ParceArgs(argc, argv, &ModFiles)==false ){
        printf("Syntax:  MikMod2 [OPTIONS] MOD_FILE [MOD_FILE...]\n");
        
        printf("Options:\n");
        printf(" -f=<5000-44100> Mixer Fequency      -p=<0-128> Set Panning Separation\n"        );
        printf(" -v=<0-128> Volume (128 is default)  -d=<1-6>   Select Driver. See List Below\n" );
        printf(" -e=<0-15>  Reverb Effect            -b=<0-3/0-7> Buffer size for OS/2 Drivers\n"   );
        printf(" -a=<#>     OS/2 Audio Device        -k  Lock the Audio Device (No Sharing)\n"     );
        printf(" -8  8-Bit Playback (16 is Default)  -q  High Quality Mixer (Slow & Buggy)\n"      );
        printf(" -i  Enable Interpolation Filter     -m  Playback in MONO (Stereo is Default)\n"   );
        printf(" -s  Enable SURROUND Sound Mode      -r  Reverse Stereo (Swap Left & Right)\n"     );
        printf(" -n  Disable internal looping        -l  Enable song looping\n"                    );
        printf(" -h  Display this Help Message       -t  Disable Time-Critical Priority\n"         );
    
        /* register all the drivers */
        MikMod_RegisterAllDrivers();

        printf( "\nDriver List:\n%s\n\n", MikMod_InfoDriver() );

        rc = MikMod_GetVersion();
        printf( "Version:   MikMod %d.%d.%d - OS/2 Interface %s  (%s %s)\n", 
                   rc>>16, (rc&0x0FFF)>>8, (rc&0x000F),
                   MIKMOD2_VER, __DATE__, __TIME__);

        return(0);
    }

    /* register all the drivers */
    MikMod_RegisterAllDrivers();

    /* register all the module loaders */
    MikMod_RegisterAllLoaders();

    /* initialize the library */
    if(MikMod_Init()) {
        fprintf(stderr,"Could not initialize sound, ErrorNo: %d  reason: %s\n", MikMod_errno, MikMod_strerror(MikMod_errno));
         return 1;
    }

    /* Set the Priority to time-critical (only if the use didn't diable this) */ 
    if( setPriority == true ){
        rc = DosSetPriority( PRTYS_PROCESS,
                                   PRTYC_TIMECRITICAL, PRTYD_MINIMUM,
                                  /* PRTYC_REGULAR, PRTYD_MAXIMUM,*/
                                  0 );
        if( rc!= 0 ){
            printf( "Error setting Priority.. Priority defaulting to the System Default.\n" );
        }
    }

    PlayModule(&ModFiles);
    
    /* give up */
    MikMod_Exit();

    return 0;
}

/************************************************************ Handle Input******/
int HandleInput(){

    int key;

     key = readKBD();
     if( key == 0 || key == 0xe0 ){ 
        key = readKBD();
        key = key<<1;
     }

//    if( key != -1 ){
//        printf( "Key = %d" , key );
//    }

    switch( key){
        case K_UP<<1:
           if( volume < 127 ) {
               volume+=8;
               Player_SetVolume(volume);
               Player_GetModule()->initvolume=volume;
           }
           break;
        case K_DOWN<<1:
            if( volume > 1 ){
                volume-=8;
                Player_SetVolume( volume ); 
                Player_GetModule()->initvolume=volume;
            }
            break;
         case K_CTRL_UP<<1:
            if( volume <127 ){
                volume++;
                Player_SetVolume(volume);
                Player_GetModule()->initvolume=volume;
            }
            break;
         case K_CTRL_DOWN<<1:
            if( volume > 1 ){
                volume--;
                Player_SetVolume( volume ); 
                Player_GetModule()->initvolume=volume;
            }
            break;
        case K_LEFT<<1:         
            Player_PrevPosition();
            break;
        case K_RIGHT<<1:
            Player_NextPosition();
            break;
         case K_HOME<<1:
            Player_SetPosition(0);
            break;
         case K_PAGEUP<<1:
            return(PREV);
            break;
         case K_PAGEDOWN<<1:
            return(NEXT);
            break;
        case 'p':
        case 'P':
            Player_TogglePause();
            break;
         case 'i':
         case 'I':
            if( Player_GetModule()->instruments==NULL ){
               DrawSamples(Player_GetModule());
            }else{
               DrawInstruments(Player_GetModule());
            }
            break;
         case 'M':
         case 'm':
            if( Player_GetModule()->comment != NULL ){
                printf( "%s\n", Player_GetModule()->comment );
            }else{
               printf( "No Message from the Author for this Module.\n" );
            }
            break;
        case 27: /* ESC */
            return(EXIT);
            break;
    }
    return(NONE);
}

/************************************************************* Parce Args ******/
int ParceArgs(int argc, char *argv[], SongList **ppList ){
    int value;
    int ret = true;
    int index;
    SongList *pList;
    int BuffSize = -1;

    *ppList = NULL;
    
    for( index=1; index<argc && ret==true ; index++ ){

        strlwr( argv[index] );

        if( argv[index][0]=='-' && (argv[index][2]=='=' || argv[index][2]=='\0') ){
            switch( argv[index][1] ){
            case 'd':
               value = atoi( &argv[index][3] );
               if( value>0 && value<7 )
                   md_device = value;
                else
                   ret = false;
               break;
            case 'f':
               value = atoi( &argv[index][3] );
               if( value>4999 && value<44101 )
                   md_mixfreq = value;
               else
                   ret = false;
               break;
            case 'v':
               value = atoi( &argv[index][3] );
               if( value>-1 && value<129 )                 
                   volume = value;
               else
                   ret = false;
               break;
            case 'e':
               value = atoi( &argv[index][3] );
               if( value>-1 && value<16 )
                  md_reverb = value;
               else
                  ret = false;
               break;
            case 'a':
               value = atoi( &argv[index][3] );
               OS2_Dart_SetOptions( "Device", value );
               OS2_MMPM_SetOptions( "Device", value );
               break;
            case 'p':
               value = atoi( &argv[index][3] );
               if( value>-1 && value<129)
                  md_pansep = value;
               else
                  ret = false;
               break;
            case 'b':
               BuffSize = atoi( &argv[index][3] );
               break;
            case '8':
               md_mode = md_mode ^ DMODE_16BITS;
               break;
            case 'q':
               md_mode = md_mode | DMODE_HQMIXER;
               break;
            case 'i':
               md_mode = md_mode | DMODE_INTERP;
               break;
            case 'm':
               md_mode = md_mode ^ DMODE_STEREO;
               break;
            case 's':
               md_mode = md_mode | DMODE_SURROUND;
               break;
            case 'r':
               md_mode = md_mode | DMODE_REVERSE;
               break;
            case 'n':
               loop = 0;
               break;
            case 'l':
               wrap = 1; 
               break;
            case 't':
               setPriority = false;
               break;
            case 'k':
               OS2_Dart_SetOptions( "NoShare", 1 );
               OS2_MMPM_SetOptions( "NoShare", 1 );
               break;
            default: 
               ret = false;
               break;
            }
        }else if( argv[index][0]!='-' ){
            // Ensure this argument is a file name.
            if( strchr( argv[index], '.') == 0 ){
                ret = false;
                break;
            }

            // Check if this File is a .lst file
            if( strcmp( (char*)strchr( argv[index], '.'), ".lst")==0 ){
                if( *ppList==NULL ){
                    pList = LoadListFile( argv[index], ppList );
                }else{
                    pList = LoadListFile( argv[index], &pList );
                }
            }else{

                if( *ppList == NULL ){
                    pList = malloc( sizeof(SongList) );
                    pList->Prev = NULL;
                    *ppList = pList;                    
                }else{
                    pList->Next = malloc( sizeof(SongList) );
                    pList->Next->Prev = pList;
                    pList = pList->Next; 
                }
                pList->Next = NULL;
                pList->pszFileName = malloc( strlen( argv[index] ) +1 );
                strcpy( pList->pszFileName, argv[index] );
            }
        }else{
            ret = false;
        }
    }

    if( *ppList == NULL ) ret = false;

    if( BuffSize != -1 ){
        if( md_device==0 || md_device == 1 ){
             if( OS2_Dart_SetOptions( "BuffSize", BuffSize+12 ) == false ) ret = false;
        }else if( md_device== 2 ){
             if( OS2_MMPM_SetOptions( "BuffSize", BuffSize+12 ) == false ) ret = false;
        }
    }

//  printf( "Returing: %d... FileName = %s\n", ret, *ppszFileName );

    return( ret );
}

int  ParceEnv(char *pszEnv){
    int   argc=0;
//    int i;
    char *pszSpace;
    char* argv[12];
    char szDummyFile[10]= "DUMMY.FIL";
    SongList *Temp;

    if( pszEnv==NULL ) return(true);
    
//    printf( "Env = %s\n", pszEnv );

    pszEnv++;   //Skip the leading char

    argv[0] = NULL;
    argv[1] = pszEnv;
    argc++;
    pszSpace = (char*)strchr(pszEnv, ' ' );
    while( pszSpace != NULL ){
        argc++;
        pszSpace[0] = '\0';
        argv[argc] = pszSpace+1;
        pszSpace = (char*)strchr(argv[argc], ' ' );    
    }

    argv[argc][strlen(argv[argc])-1]='\0';  // Remove the trailing char

    argc++;
    argv[argc] =  szDummyFile;
    argc++;
        
//    for(i=0;i<argc;i++) {
//        printf( "Arg%d = %s\n", i, argv[i] );
//    }
  
    return( ParceArgs( argc, argv, &Temp) );
}

void DrawInstruments(MODULE *module){
    unsigned short cnt;
    short              middle;
    short              odd;

//    printf( "NumIns = %d, NumSmp = %d\n" , module->numins, module->numsmp );
    odd    = (short)module->numins % 2;
    middle = (short)module->numins / 2;    

    printf( " Ŀ\n" ); 

    for( cnt=0; cnt < middle ; cnt++ ){
    
        if( module->instruments==NULL ) return;

        if( module->instruments[cnt].insname == NULL ){
            printf( "                                     " );
        }else{
            printf( "  %-35s ", module->instruments[cnt].insname );
        }
        if( cnt < middle-1 || odd==0 ){
           if( module->instruments[middle+cnt].insname == NULL ){
               printf( "                                     \n" );
           }else{
               printf( " %-35s \n", module->instruments[middle+cnt].insname );
           }
        }else{
            printf( "                                     \n" );
        }
    }
    printf( " \n" ); 
}



void DrawSamples(MODULE *module){
    unsigned short     cnt;
    short              middle;
    short              odd;

//    printf( "NumIns = %d, NumSmp = %d\n" , module->numins, module->numsmp );
    odd    = (short)module->numsmp % 2;
    middle = (short)module->numsmp / 2;    

    printf( " Ŀ\n" ); 

    for( cnt=0; cnt < middle ; cnt++ ){
    
        if( module->samples==NULL ) return;

        if( module->samples[cnt].samplename == NULL ){
            printf( "                                      " );
        }else{
            printf( "  %-35s ", module->samples[cnt].samplename );
        }
        if( cnt < middle-1 || odd==0 ){
           if( module->samples[middle+cnt].samplename == NULL ){
               printf( "                                     \n" );
           }else{
               printf( " %-35s \n", module->samples[middle+cnt].samplename );
           }
        }else{
            printf( "                                     \n" );
        }
    }
    printf( " \n" ); 
}


/************************************************************* Play Module *****/
int PlayModule(SongList **ppList){

    SongList *pList;
    MODULE *module;
    int cmd = NONE;

    pList = *ppList;

    while( pList!=NULL && cmd!=EXIT ){

        /* load module */
        module = Player_Load( pList->pszFileName,128,0);
        if (module) {

            /* Print  some Neat Info */
            printf( "\n" );
            printf( " Ŀ\n" );
            printf( "  Volume:   Rwd/Fwd:   Restart: <HOME>  Instruments: I  Exit: <ESC> \n" );
            printf( " Ĵ\n" );
            printf( "  Message: M  Pause: P  Next Song: <PAGE-DOWN>  Previous Song: <PAGE-UP> \n" );
            printf( " Ĵ\n" );
            printf( "    Module: %-64s\n", pList->pszFileName  );
            printf( " Ĵ\n" );
            printf( "     Title: %-63s \n", module->songname );
            printf( " Ĵ\n" );
            printf( "  MOD Type: %-47s  Channels: %-3d \n", module->modtype, module->numchn  );
            printf( " \n" );
            module->volume = volume;
            module->initvolume = volume;

            /* start module */
            Player_Start(module);

            /* Setup this module before playing */
            module->loop = loop;
            module->wrap = wrap;
     
            cmd = NONE;

            while( Player_Active() && cmd==NONE ) {
                /* we're playing */
                MikMod_Update();
                cmd = HandleInput();
                if( (cmd==NEXT&&pList->Next==NULL) || (cmd==PREV&&pList->Prev==NULL) ){
                   cmd=NONE;
                }
                DosSleep(50);
            }
     
            Player_Stop();
            Player_Free(module);
            if( md_device!=2 ) MikMod_Reset();
        } else {
             fprintf(stderr,"Could not load module, ErroNo: %d  reason: %s\n",  MikMod_errno, MikMod_strerror(MikMod_errno));    
        }     
        if( cmd==PREV ){
            pList = pList->Prev;
        }else{
            pList = pList->Next;
        }
    }
}


/************************************************************* Load List File *****/
SongList* LoadListFile(char *pszFileName, SongList **ppList){
    SongList *pList;
    FILE     *file;
    int      iLen;
    char     szFileName[256];

    pList = *ppList;

    file = fopen( pszFileName, "rt" );
    if( file!=NULL ){
        while( fgets( szFileName, 255, file ) != NULL ){
            if(*ppList==NULL){
                *ppList = malloc( sizeof(SongList) );
                pList = *ppList;
                pList->Next = NULL;
                pList->Prev = NULL;
            }else{
                pList->Next = malloc( sizeof(SongList) );
                pList->Next->Prev = pList;
                pList = pList->Next;
                pList->Next = NULL;
            }
            iLen = strlen( szFileName );
            iLen--;
            while( szFileName[iLen]== '\n' ) { szFileName[iLen]='\0'; }
            pList->pszFileName = malloc( strlen( szFileName ) +1 );
            strcpy( pList->pszFileName, szFileName );
        }
    }
    return( pList );
}
